#pragma once

#include "teqp/models/multifluid.hpp"
#include <Eigen/Dense>
#include <math.h>
#include "teqp/models/cubics.hpp"
#include <algorithm>
#include <cctype>
#include <iostream>
#include <string>
#include <vector>

namespace teqp {
    namespace Mie {


        /**
               Equation of state for the Mie ($\lambda_{\mathrm{r}}$,6) fluid with a repulsive exponent from 11 to 13
               Pohl, S.; Fingerhut, R., Thol, M.; Vrabec, J.; Span, R.
               */

        enum combining_rule { ONEFLUID, ONEFLUID_RED, ONEFLUID_RED_LINEAR, ONEFLUID_DENSE, LINEAR, LORENTZ };

        class Mie6Pohl2023 {

        private:
            using EArray6 = Eigen::Array<double, 6, 1>;
            using EArray4 = Eigen::Array<double, 4, 1>;

            const EArray6 c1_pol = (EArray6() << -0.0192944, 1.38, -2.2653, 1.6291, -1.974, 0.40412).finished();
            const EArray6 c1_exp = (EArray6() << 0.1845, -0.3227, 1.1351, 2.232, -2.344, -0.4238).finished();
            const EArray4 c1_gbs = (EArray4() << -4.367, 0.0371, 1.3895, 2.835).finished();
            const EArray6 c2_pol = (EArray6() << 0.26021, -5.525, 8.329, -19.492, 25.8, -3.8133).finished();
            const EArray6 c2_exp = (EArray6() << -5.05, 2.7842, -9.523, -30.383, 17.902, 2.2264).finished();
            const EArray4 c2_gbs = (EArray4() << 48.445, -5.506, -11.643, -24.36).finished();
            const EArray6 t_pol = (EArray6() << 1, 0.236, 0.872, 0.313, 0.407, 0.703).finished();
            const EArray6 t_exp = (EArray6() << 1.78, 2.99, 2.866, 1.2, 3.06, 1.073).finished();
            const EArray4 t_gbs = (EArray4() << 1.50, 1.03, 4.02, 1.57).finished();
            const EArray6 d_pol = (EArray6() << 4, 1, 1, 2, 2, 3).finished();
            const EArray6 d_exp = (EArray6() << 1, 1, 3, 2, 2, 5).finished();
            const EArray4 d_gbs = (EArray4() << 2, 3, 2, 2).finished();
            const EArray6 p = (EArray6() << 1, 2, 2, 1, 2, 1).finished();
            const EArray4 eta = (EArray4() << 0.362, 0.313, 1.17, 0.957).finished();
            const EArray4 beta = (EArray4() << 0.0761, 0.143, 0.63, 1.32).finished();
            const EArray4 gam = (EArray4() << 1.55, -0.0826, 1.505, 1.07).finished();
            const EArray4 eps = (EArray4() << -1, -1, -0.195, -0.287).finished();

            const double m_lambda_a;
            const EArray6 n_pol, n_exp;
            const EArray4 n_gbs;
            const double Tc, rhoc; // In simulation units
        public:

            Mie6Pohl2023(double lambda_a) : m_lambda_a(lambda_a),
                n_pol(c1_pol + c2_pol / m_lambda_a),
                n_exp(c1_exp + c2_exp / m_lambda_a),
                n_gbs(c1_gbs + c2_gbs / m_lambda_a),
                Tc(0.668 + 6.84 / m_lambda_a + 145 / pow(m_lambda_a, 3)), // T^*
                rhoc(0.2516 + 0.049 * log10(m_lambda_a)) // rho^*
            {}

            auto get_lambda_a() const { return m_lambda_a; }

            // We are in "simulation units", so R is 1.0, and T and rho that
            // go into alphar are actually T^* and rho^*
            template<typename MoleFracType>
            double R(const MoleFracType&) const { return 1.0; }

            template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
            auto alphar(const TTYPE& Tstar, const RHOTYPE& rhostar, const MoleFracType& /*molefrac*/) const {
                auto tau = Tc / Tstar; auto delta = rhostar / rhoc;
                auto alphar_ = (
                    (n_pol * pow(tau, t_pol) * pow(delta, d_pol)).sum() +
                    (n_exp * pow(tau, t_exp) * pow(delta, d_exp) * exp(-pow(delta, p))).sum() +
                    (n_gbs * pow(tau, t_gbs) * pow(delta, d_gbs) * exp(-eta * pow(delta - eps, 2) - beta * pow(tau - gam, 2))).sum()
                    );
                return forceeval(alphar_);
            }

        };


        //
        inline auto linear_mixing(const double& x, const double& y) {
            return (x + y) / 2.0;
        }

        class AncTerm {
        public:
            Eigen::ArrayXd n, t;
            double tred, dred;
            std::string type;
        };

        struct fluid {
            double lambdas, epsilons, sigmas, m, I;
            AncTerm anc_dl, anc_dv;
            bool is_sphere;
        };


        // Combining rules for one fluid approximation
        // (1) Simple Van der Waals one fluid combininb rule
        template<typename RHOTYPE, typename MoleFracType>
        inline auto combining_rules_one_fluid(RHOTYPE& rhostar, MoleFracType& molefrac, std::vector<fluid> f, const bool& is_spherical, Eigen::ArrayXXd k_mat_l, Eigen::ArrayXXd k_mat_s, Eigen::ArrayXXd k_mat_e, Eigen::ArrayXd gammat, Eigen::ArrayXd betat, Eigen::ArrayXd gamma_dense, combining_rule mix_rule) {
            auto ncomp = f.size();
            using resulttype = std::common_type_t<decltype(molefrac[0]), decltype(rhostar)>;
            using rhotype = std::common_type_t<decltype(rhostar)>;
            std::vector<std::vector<resulttype>> sigma_ij(ncomp, std::vector<resulttype>(ncomp, 0.0));
            std::vector<std::vector<resulttype>> eps_ij(ncomp, std::vector<resulttype>(ncomp, 0.0));
            std::vector<std::vector<resulttype>> lambda_ij(ncomp, std::vector<resulttype>(ncomp, 0.0));
            std::vector<resulttype> m_mixed(ncomp, 0.0);
            resulttype m_mix = 0.0;
            resulttype sigma_mean = 0.0;
            resulttype sigma_mean_gamma = 0.0;
            resulttype epsilon_mean = 0.0;
            resulttype lambda_mean = 0.0;


            constexpr double MY_PI = EIGEN_PI;
            double pi6 = (MY_PI / 6.0);
            resulttype rho_pack = pi6 * rhostar * N_A * 1e-30;
            resulttype chsi = 0.0;
            resulttype gamma_exp = 0.0;

            if (is_spherical) {
                m_mix = 1.0;
            }
            else
            {
                for (auto i = 0; i < ncomp; i++) {
                    m_mix += f[i].m * molefrac[i];
                }
            }

            //resulttype kij = 0.00;
            //for (size_t i = 0; i < ncomp - 1; i++)
            //{
            //    //if(molefrac[0]<1.0) {
            //    kij = 0.0;//1.0  - pow(2.0, 7.0) * (sqrt(f[i].I * f[i + 1].I) / (f[i].I + f[i + 1].I)) * (pow(f[i].sigmas, 3.0) * pow(f[i + 1].sigmas, 3.0)) / pow(f[i].sigmas + f[i + 1].sigmas, 6);
            //}
            Eigen::ArrayXd sig_;
            Eigen::ArrayXd lambda_;
            Eigen::ArrayXd epsilon_;

            auto red_func = [molefrac](auto p, auto b, auto g) {
                resulttype val = 0.0;
                for (size_t i = 0; i < molefrac.size(); i++) {
                    val = val + molefrac[i] * molefrac[i] * p(i);
                }
                val = val + 2.0 * molefrac[0] * molefrac[1] * b * g * (molefrac[0] + molefrac[1]) / (b * b * molefrac[0] + molefrac[1]) * sqrt(p(0) * p(1));
                return val;
            };



            auto red_func_lambda = [molefrac](auto p, auto b, auto g) {
                resulttype val = 0.0;
                for (size_t i = 0; i < molefrac.size(); i++) {
                    val = val + molefrac[i] * molefrac[i] * p(i);
                }
                val = val + 2.0 * molefrac[0] * molefrac[1] * b * g * (molefrac[0] + molefrac[1]) / (b * b * molefrac[0] + molefrac[1]) * (sqrt((p(0) - 3.0) * (p(1) - 3.0)) + 3.0);
                return val;
            };





            auto red_func_d = [molefrac](auto p, auto b, auto g) {
                resulttype val = 0.0;
                for (size_t i = 0; i < molefrac.size(); i++) {
                    val = val + molefrac[i] * molefrac[i] / p(i);
                }
                val = val + 2.0 * molefrac[0] * molefrac[1] * b * g * (molefrac[0] + molefrac[1]) / (b * b * molefrac[0] + molefrac[1]) * 0.125 * pow(1.0 / pow(p(0), 1.0 / 3.0) + 1.0 / pow(p(1), 1.0 / 3.0), 3.0);

                return val;
            };

            // Switch between combing rules for interaction of molecular parameters
            switch (mix_rule) {
            case ONEFLUID:
                if (ncomp > 1) {
                    for (auto i = 0; i < ncomp; i++) {
                        for (auto j = 0; j < ncomp; j++) {
                            //lambda_ij[i][j] = (1.0 - k_mat_l(i, j)) * linear_mixing(f[i].lambdas, f[j].lambdas); //3.0 + sqrt((f[i].lambdas - 3.0) * (f[j].lambdas - 3.0));
                            lambda_ij[i][j] = (1.0 - k_mat_l(i, j)) * sqrt((f[i].lambdas - 3.0) * (f[j].lambdas - 3.0)) + 3.0;
                            sigma_ij[i][j] = (1.0 - k_mat_s(i, j)) * linear_mixing(f[i].sigmas, f[j].sigmas);
                            eps_ij[i][j] = (1.0 - k_mat_e(i, j)) * sqrt(pow(f[i].sigmas, 3.0) * pow(f[j].sigmas, 3.0)) / pow(sigma_ij[i][j], 3.0) * sqrt(f[i].epsilons * f[j].epsilons);//*pow(sqrt(f[i].sigmas*f[j].sigmas)/ sigma_ij[i][j],6.0); //sqrt(pow(f[i].sigmas,3.0)*pow(f[j].sigmas,3.0))/pow(sigma_ij[i][j],3.0)
                        }
                    }
                    // Calculate the mean values for the one fluid approximation
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { sigma_mean += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 3.0); } }
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { epsilon_mean += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 3.0) * eps_ij[i][j]; } }
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { lambda_mean += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 3.0) * eps_ij[i][j] * lambda_ij[i][j]; } }
                    /*                   for (auto i = 0; i < ncomp; i++) {
                                           lambda_mean += molefrac[i] * f[i].lambdas;
                                       }
                                       for (auto i = 0; i < ncomp; i++) {
                                           epsilon_mean += molefrac[i] * f[i].epsilons;
                                       }
                                       for (auto i = 0; i < ncomp; i++) {
                                           sigma_mean += molefrac[i] * f[i].sigmas;
                                       }*/
                    lambda_mean = lambda_mean / epsilon_mean;
                    epsilon_mean = epsilon_mean / sigma_mean;
                    sigma_mean = pow(sigma_mean / pow(m_mix, 2.0), 1.0 / 3.0);

                }
                else
                {
                    sigma_mean = f[0].sigmas;
                    epsilon_mean = f[0].epsilons;
                    lambda_mean = f[0].lambdas;
                }
                break;
            case ONEFLUID_DENSE:




                if (ncomp > 1) {

                    for (auto i = 0; i < ncomp; i++) {
                        chsi += molefrac[i] * pow(f[i].sigmas, 3.0);
                    }
                    chsi = chsi * rho_pack;
                    gamma_exp = 1.0 / 3.0 + 2.0 / 3.0 * chsi;

                    for (auto i = 0; i < ncomp; i++) {
                        for (auto j = 0; j < ncomp; j++) {
                            lambda_ij[i][j] = (1.0 - k_mat_l(i, j)) * linear_mixing(f[i].lambdas, f[j].lambdas); //3.0 + sqrt((f[i].lambdas - 3.0) * (f[j].lambdas - 3.0));
                            sigma_ij[i][j] = (1.0 - k_mat_s(i, j)) * pow((pow(f[i].sigmas, 3.0 * gamma_exp) + pow(f[j].sigmas, 3.0 * gamma_exp)) / 2.0, 1.0 / gamma_exp);//linear_mixing(f[i].sigmas, f[j].sigmas);
                            eps_ij[i][j] = (1.0 - k_mat_e(i, j)) * sqrt(f[i].epsilons * f[j].epsilons);//*pow(sqrt(f[i].sigmas*f[j].sigmas)/ sigma_ij[i][j],6.0); //sqrt(pow(f[i].sigmas,3.0)*pow(f[j].sigmas,3.0))/pow(sigma_ij[i][j],3.0)
                        }
                    }



                    // Calculate the mean values for the one fluid approximation
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { sigma_mean_gamma += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 1.0); } }
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { epsilon_mean += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 1.0) * eps_ij[i][j]; } }
                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { sigma_mean += molefrac[i] * molefrac[j] * f[i].m * f[j].m * pow(sigma_ij[i][j], 1.0); } }

                    for (auto i = 0; i < ncomp; i++) { for (auto j = 0; j < ncomp; j++) { lambda_mean += molefrac[i] * molefrac[j] * lambda_ij[i][j]; } }
                    epsilon_mean = epsilon_mean / sigma_mean_gamma;
                    sigma_mean = pow(sigma_mean / pow(m_mix, 2.0), 1.0 / 3.0);
                }
                else
                {
                    sigma_mean = f[0].sigmas;
                    epsilon_mean = f[0].epsilons;
                    lambda_mean = f[0].lambdas;
                }

                break;

            case ONEFLUID_RED:
                // so far only for binary mixtures available!
                if (ncomp > 1) {
                    sig_ = (Eigen::ArrayXd(2) << f[0].sigmas, f[1].sigmas).finished();
                    lambda_ = (Eigen::ArrayXd(2) << f[0].lambdas, f[1].lambdas).finished();
                    epsilon_ = (Eigen::ArrayXd(2) << f[0].epsilons, f[1].epsilons).finished();
                    sigma_mean = 1.0 / red_func_d(sig_, betat[0], gammat[0]);
                    epsilon_mean = red_func(epsilon_, betat[1], gammat[1]);
                    lambda_mean = red_func(lambda_, betat[2], gammat[2]);
                }
                else
                {
                    sigma_mean = f[0].sigmas;
                    epsilon_mean = f[0].epsilons;
                    lambda_mean = f[0].lambdas;
                }
                break;
            case ONEFLUID_RED_LINEAR:
                // so far only for binary mixtures available!
                if (ncomp > 1) {
                    sig_ = (Eigen::ArrayXd(2) << f[0].sigmas, f[1].sigmas).finished();
                    lambda_ = (Eigen::ArrayXd(2) << f[0].lambdas, f[1].lambdas).finished();
                    epsilon_ = (Eigen::ArrayXd(2) << f[0].epsilons, f[1].epsilons).finished();
                    sigma_mean = 1.0 / red_func_d(sig_, 1.0, gammat[0]);
                    epsilon_mean = red_func(epsilon_, 1.0, gammat[1]);
                    lambda_mean = red_func(lambda_, 1.0, gammat[2]);
                }
                else
                {
                    sigma_mean = f[0].sigmas;
                    epsilon_mean = f[0].epsilons;
                    lambda_mean = f[0].lambdas;
                }
                break;
            default:
                break;
            }

            return std::make_tuple(sigma_mean, epsilon_mean, lambda_mean, m_mix);
        }


        inline auto dlv_eq1(const double& T, const AncTerm& anc) {
            auto tau = abs(1.0 - T / anc.tred);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * (sum + 1.0);
        }

        inline auto dlv_eq2(const double& T, const AncTerm& anc) {
            auto tau = pow(abs(1.0 - T / anc.tred), 1.0 / 3.0);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * (sum + 1.0);
        }

        inline auto dlv_eq3(const double& T, const AncTerm& anc) {
            auto tau = abs(1.0 - T / anc.tred);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * exp(sum);
        }

        inline auto dlv_eq4(const double& T, const AncTerm& anc) {
            auto tau = pow(abs(1.0 - T / anc.tred), 1.0 / 3.0);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * exp(sum);
        }

        inline auto dlv_eq5(const double& T, const AncTerm& anc) {
            auto tau = abs(1.0 - T / anc.tred);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * exp(anc.tred / T * sum);
        }

        inline auto dlv_eq6(const double& T, const AncTerm& anc) {
            auto tau = pow(abs(1.0 - T / anc.tred), 1.0 / 3.0);
            auto sum = 0.0;
            for (size_t i = 0; i < anc.n.size(); i++) {
                sum = sum + anc.n(i) * pow(tau, anc.t(i));
            }
            return anc.dred * exp(anc.tred / T * sum);
        }



        template<typename Model>
        auto get_vle_pure_start(const Model& model, double& T, const int idx) {
            return get_sat_dense(T, model, idx);
        }

        const  double kBoltz = 1.380649E-23;
        const double NAvo = 6.02214076E23;

        class MieElong {

        private:

            const  double kBoltz = 1.380649E-23;
            const double NAvo = 6.02214076E23;
            using EArray6 = Eigen::Array<double, 6, 1>;
            using EArray4 = Eigen::Array<double, 4, 1>;


            Eigen::ArrayXd c1_pol;
            Eigen::ArrayXd c1_exp;
            Eigen::ArrayXd c1_gbs;
            Eigen::ArrayXd c2_pol;
            Eigen::ArrayXd c2_exp;
            Eigen::ArrayXd c2_gbs;
            Eigen::ArrayXd t_pol;
            Eigen::ArrayXd t_exp;
            Eigen::ArrayXd t_gbs;
            Eigen::ArrayXd d_pol;
            Eigen::ArrayXd d_exp;
            Eigen::ArrayXd d_gbs;
            Eigen::ArrayXd p;
            Eigen::ArrayXd eta;
            Eigen::ArrayXd beta;
            Eigen::ArrayXd gam;
            Eigen::ArrayXd eps;
            Eigen::ArrayXd tc_p;
            Eigen::ArrayXd dc_p;
            Eigen::ArrayXd d1_pol;
            Eigen::ArrayXd d2_pol;
            Eigen::ArrayXd d3_pol;
            Eigen::ArrayXd d1_exp;
            Eigen::ArrayXd d2_exp;
            Eigen::ArrayXd d3_exp;
            Eigen::ArrayXd d1_gbs;
            Eigen::ArrayXd d2_gbs;
            Eigen::ArrayXd d3_gbs;
            Eigen::ArrayXd t_pol_m;
            Eigen::ArrayXd t_exp_m;
            Eigen::ArrayXd t_gbs_m;
            Eigen::ArrayXd d_pol_m;
            Eigen::ArrayXd d_exp_m;
            Eigen::ArrayXd d_gbs_m;
            Eigen::ArrayXd p_m;
            Eigen::ArrayXd eta_m;
            Eigen::ArrayXd beta_m;
            Eigen::ArrayXd gam_m;
            Eigen::ArrayXd eps_m;

            int nr_fld;
            bool is_sphere;

            Eigen::ArrayXd kij_vec, betat, gammat, gamma_dense;
            Eigen::ArrayXXd k_mat_lambda, k_mat_sigma, k_mat_epsilon;
            combining_rule mix_rule;
            std::map<std::string, combining_rule>  comb_rule = { {"one-fluid",ONEFLUID} , {"one-fluid-red",ONEFLUID_RED} , {"one-fluid-linear",ONEFLUID_RED_LINEAR}, {"linear",LINEAR} , {"lorentz-berthelot",LORENTZ} , {"one-fluid-dense",ONEFLUID_DENSE} };
        public:

            std::vector<fluid> fld;
            //std::string combining_rule;

            //MieElong(double lambda_a, double ms, double sigma, double eps) { // Repulsive exponent, segment number ms
            MieElong(const std::string& path, const std::vector<std::string>& fluids, const std::string& combining_rule_in, const std::vector<double>& kij = {}) {

                std::string filepath = std::filesystem::is_regular_file(path) ? path : path;
                nlohmann::json j = load_a_JSON_file(filepath);
                std::string combining_rule = combining_rule_in;
                auto nr_fluids = fluids.size();
                std::transform(combining_rule.begin(), combining_rule.end(), combining_rule.begin(), [](unsigned char c) { return std::tolower(c); });
                

                try {
                    // Attempt to access the map with the user's input
                    int value = comb_rule.at(combining_rule);
                    mix_rule = comb_rule[combining_rule];
                }
                catch (const std::out_of_range& e) {
                    // Handle the exception when the key is not found
                    std::cerr << "Error: " << e.what() << " is not a valid mixing rule." << std::endl;
                }

                fld.resize(fluids.size());
                    int i = 0;
                for (auto f : fluids)
                {
                    fld[i].lambdas = j.at("mie").at(f).at("lambda");
                    fld[i].epsilons = j.at("mie").at(f).at("epsilon");
                    fld[i].sigmas = j.at("mie").at(f).at("sigma");
                    fld[i].m = j.at("mie").at(f).at("segment");
                    fld[i].I = j.at("mie").at(f).at("I");

                    auto n_anc = static_cast<int>(j.at("mie").at(f).at("dl_coeff").size());
                    fld[i].anc_dl.tred = j.at("mie").at(f).at("tred");
                    fld[i].anc_dl.dred = j.at("mie").at(f).at("dred");
                    fld[i].anc_dl.type = j.at("mie").at(f).at("dl_type");
                    fld[i].anc_dl.n = toeig(j.at("mie").at(f).at("dl_coeff")).head(n_anc);
                    fld[i].anc_dl.t = toeig(j.at("mie").at(f).at("dl_texp")).head(n_anc);

                    n_anc = static_cast<int>(j.at("mie").at(f).at("dv_coeff").size());
                    fld[i].anc_dv.tred = j.at("mie").at(f).at("tred");
                    fld[i].anc_dv.dred = j.at("mie").at(f).at("dred");
                    fld[i].anc_dv.type = j.at("mie").at(f).at("dv_type");
                    fld[i].anc_dv.n = toeig(j.at("mie").at(f).at("dv_coeff")).head(n_anc);
                    fld[i].anc_dv.t = toeig(j.at("mie").at(f).at("dv_texp")).head(n_anc);

                    i++;
                }

                nr_fld = fluids.size();
                is_sphere = true;
                /*combining_rule = combining_rule_in;*/
                // check if only sphericals are included
                for (size_t i = 0; i < nr_fld; i++)
                {
                    if (fld[i].m > 1.0) {
                        is_sphere = false;
                    }
                }

                // get mie parameter
                auto spec = j.at("mie").at("parameter");
                auto n_pol = static_cast<int>(spec.at("c1_pol").size());
                auto n_exp = static_cast<int>(spec.at("c1_exp").size());
                auto n_gbs = static_cast<int>(spec.at("c1_gbs").size());
                c1_pol = toeig(spec.at("c1_pol")).head(n_pol);
                c1_exp = toeig(spec.at("c1_exp")).head(n_exp);
                c1_gbs = toeig(spec.at("c1_gbs")).head(n_gbs);
                c2_pol = toeig(spec.at("c2_pol")).head(n_pol);
                c2_exp = toeig(spec.at("c2_exp")).head(n_exp);
                c2_gbs = toeig(spec.at("c2_gbs")).head(n_gbs);
                t_pol = toeig(spec.at("t_pol")).head(n_pol);
                t_exp = toeig(spec.at("t_exp")).head(n_exp);
                t_gbs = toeig(spec.at("t_gbs")).head(n_gbs);
                d_pol = toeig(spec.at("d_pol")).head(n_pol);
                d_exp = toeig(spec.at("d_exp")).head(n_exp);
                d_gbs = toeig(spec.at("d_gbs")).head(n_gbs);
                p = toeig(spec.at("p")).head(n_exp);
                eta = toeig(spec.at("eta")).head(n_gbs);
                beta = toeig(spec.at("beta")).head(n_gbs);
                gam = toeig(spec.at("gam")).head(n_gbs);
                eps = toeig(spec.at("eps")).head(n_gbs);

                n_pol = static_cast<int>(spec.at("d1_pol").size());
                n_exp = static_cast<int>(spec.at("d1_exp").size());
                n_gbs = static_cast<int>(spec.at("d1_gbs").size());
                d1_pol = toeig(spec.at("d1_pol")).head(n_pol);
                d1_exp = toeig(spec.at("d1_exp")).head(n_exp);
                d1_gbs = toeig(spec.at("d1_gbs")).head(n_gbs);
                d2_pol = toeig(spec.at("d2_pol")).head(n_pol);
                d2_exp = toeig(spec.at("d2_exp")).head(n_exp);
                d2_gbs = toeig(spec.at("d2_gbs")).head(n_gbs);
                d3_pol = toeig(spec.at("d3_pol")).head(n_pol);
                d3_exp = toeig(spec.at("d3_exp")).head(n_exp);
                d3_gbs = toeig(spec.at("d3_gbs")).head(n_gbs);
                t_pol_m = toeig(spec.at("t_pol_m")).head(n_pol);
                t_exp_m = toeig(spec.at("t_exp_m")).head(n_exp);
                t_gbs_m = toeig(spec.at("t_gbs_m")).head(n_gbs);
                d_pol_m = toeig(spec.at("d_pol_m")).head(n_pol);
                d_exp_m = toeig(spec.at("d_exp_m")).head(n_exp);
                d_gbs_m = toeig(spec.at("d_gbs_m")).head(n_gbs);
                p_m = toeig(spec.at("p_m")).head(n_exp);
                eta_m = toeig(spec.at("eta_m")).head(n_gbs);
                beta_m = toeig(spec.at("beta_m")).head(n_gbs);
                gam_m = toeig(spec.at("gam_m")).head(n_gbs);
                eps_m = toeig(spec.at("eps_m")).head(n_gbs);

                auto n_tc = static_cast<int>(spec.at("tc_p").size());
                auto n_dc = static_cast<int>(spec.at("dc_p").size());
                tc_p = toeig(spec.at("tc_p")).head(n_tc);
                dc_p = toeig(spec.at("dc_p")).head(n_dc);


                k_mat_lambda = Eigen::ArrayXXd::Zero(nr_fluids, nr_fluids);
                k_mat_sigma = Eigen::ArrayXXd::Zero(nr_fluids, nr_fluids);
                k_mat_epsilon = Eigen::ArrayXXd::Zero(nr_fluids, nr_fluids);

                // Binary interaction parameter
                if (kij.size() == 0) {
                    // Check if a mixture is present
                    if (nr_fluids > 1) {
                        if (mix_rule == ONEFLUID || mix_rule == ONEFLUID_DENSE) {
                            auto spec_mix = j.at("kij");
                            // Find fluids
                            for (size_t i = 0; i < nr_fluids - 1; i++) {
                                for (size_t k = i + 1; k < nr_fluids; k++) {
                                    // build name combinations
                                    std::string name1 = fluids[i] + "-" + fluids[k];
                                    std::string name2 = fluids[k] + "-" + fluids[i];
                                    if (spec_mix.contains(name1)) {
                                        //kij_vec = toeig(spec_mix.at(name1)).head(3);
                                        auto kmat_loc = spec_mix.at(name1);
                                        k_mat_lambda(i, k) = kmat_loc[0];
                                        k_mat_lambda(k, i) = kmat_loc[0];
                                        k_mat_sigma(i, k) = kmat_loc[1];
                                        k_mat_sigma(k, i) = kmat_loc[1];
                                        k_mat_epsilon(i, k) = kmat_loc[2];
                                        k_mat_epsilon(k, i) = kmat_loc[2];
                                    }
                                    else if (spec_mix.contains(name2)) {
                                        //kij_vec = toeig(spec_mix.at(name2)).head(3);
                                        auto kmat_loc = spec_mix.at(name2);
                                        k_mat_lambda(i, k) = kmat_loc[0];
                                        k_mat_lambda(k, i) = kmat_loc[0];
                                        k_mat_sigma(i, k) = kmat_loc[1];
                                        k_mat_sigma(k, i) = kmat_loc[1];
                                        k_mat_epsilon(i, k) = kmat_loc[2];
                                        k_mat_epsilon(k, i) = kmat_loc[2];
                                    }
                                }
                            }
                        }
                        else if (mix_rule == ONEFLUID_RED) {
                            auto spec_mix = j.at("one_fluid_red");
                            for (size_t i = 0; i < nr_fluids - 1; i++) {
                                for (size_t k = i + 1; k < nr_fluids; k++) {
                                    // build name combinations
                                    std::string name1 = fluids[i] + "-" + fluids[k];
                                    std::string name2 = fluids[k] + "-" + fluids[i];
                                    if (spec_mix.contains(name1)) {
                                        auto parameter = spec_mix.at(name1);
                                        betat = toeig(parameter.at("betat")).head(3);
                                        gammat = toeig(parameter.at("gammat")).head(3);
                                    }
                                    else if (spec_mix.contains(name2)) {
                                        auto parameter = spec_mix.at(name2);
                                        betat = toeig(parameter.at("betat")).head(3);
                                        gammat = toeig(parameter.at("gammat")).head(3);
                                    }
                                }
                            }
                        }
                        else if (mix_rule == ONEFLUID_RED_LINEAR) {
                            gammat = (Eigen::ArrayXd(3) << 0.0, 0.0, 0.0).finished();
                            for (size_t i = 0; i < nr_fluids - 1; i++) {
                                for (size_t k = i + 1; k < nr_fluids; k++) {
                                    gammat(0) = 4.0 * ( 1.0/fld[i].sigmas + 1.0 / fld[k].sigmas) / pow(((1.0/pow(fld[i].sigmas,1.0/3.0))+ (1.0 / pow(fld[k].sigmas, 1.0 / 3.0))),3.0);
                                    gammat(1) = 0.5 * (fld[i].epsilons+fld[k].epsilons)/sqrt(fld[i].epsilons * fld[k].epsilons);
                                    gammat(2) = 0.5 * (fld[i].lambdas + fld[k].lambdas) / sqrt(fld[i].lambdas * fld[k].lambdas);
                                }
                            }
                        }

                        //else if (mix_rule == ONEFLUID_DENSE) {
                        //    auto spec_mix = j.at("one_fluid_dense");
                        //    for (size_t i = 0; i < nr_fluids - 1; i++) {
                        //        for (size_t k = i + 1; k < nr_fluids; k++) {
                        //            // build name combinations
                        //            std::string name1 = fluids[i] + "-" + fluids[k];
                        //            std::string name2 = fluids[k] + "-" + fluids[i];
                        //            if (spec_mix.contains(name1)) {
                        //                auto parameter = spec_mix.at(name1);
                        //                gamma_dense = (Eigen::ArrayXd(1) << parameter.at("gamma_dense")).finished();
                        //        
                        //            }
                        //            else if (spec_mix.contains(name2)) {
                        //                auto parameter = spec_mix.at(name2);
                        //                gamma_dense = (Eigen::ArrayXd(1) << parameter.at("gamma_dense")).finished();
                        //            }
                        //        }
                        //    }
                        //}

                    }
                    else {
                        kij_vec = (Eigen::ArrayXd(3) << 0.0, 0.0, 0.0).finished();
                    }
                }
                else
                {
                    if (mix_rule == ONEFLUID || mix_rule == ONEFLUID_DENSE) {
                        for (size_t i = 0; i < nr_fluids - 1; i++) {
                            for (size_t k = i + 1; k < nr_fluids; k++) {
                                k_mat_lambda(i, k) = kij[0];
                                k_mat_lambda(k, i) = kij[0];
                                k_mat_sigma(i, k) = kij[1];
                                k_mat_sigma(k, i) = kij[1];
                                k_mat_epsilon(i, k) = kij[2];
                                k_mat_epsilon(k, i) = kij[2];
                            }
                        }
                        //kij_vec = toeig(kij);
                    }
                    else if (mix_rule == ONEFLUID_RED) {
                        betat = (Eigen::ArrayXd(3) << kij[0], kij[1], kij[2]).finished();
                        gammat = (Eigen::ArrayXd(3) << kij[3], kij[4], kij[5]).finished();
                    }
                    else if (mix_rule == ONEFLUID_DENSE) {
                        gamma_dense = (Eigen::ArrayXd(1) << kij[0]).finished();
                    }
                }
            }

            template<typename MoleFracType>
            auto R(const MoleFracType&) const { return NAvo * kBoltz; }

            template<typename TTYPE, typename MoleFracType, typename GammaType>
            inline auto get_tred_mix(const TTYPE& tc, const MoleFracType& x, const GammaType& gama_t) const {
                typename TTYPE::value_type tred_mix = 0.0;
                for (size_t i = 0; i < x.size(); i++) {
                    tred_mix += x[i] * x[i] * tc[i];
                }
                auto beta_t = 1.0;
                for (size_t i = 0; i < x.size() - 1; i++) {
                    for (size_t j = i + 1; j < x.size(); j++) {
                        tred_mix += 2.0 * x[i] * x[j] * beta_t * gama_t[i][j] * (x[i] + x[j]) / (beta_t * beta_t * x[i] + x[j]) * sqrt(tc[i] * tc[j]);
                    }
                }
                return tred_mix;
            }

            template<typename TTYPE, typename MoleFracType, typename GammaType>
            inline auto get_vred_mix(const TTYPE& dc, const MoleFracType& x, const GammaType& gama_v) const {
                typename TTYPE::value_type vred_mix1 = 0.0;
                for (size_t i = 0; i < x.size(); i++) {
                    vred_mix1 += (x[i] * x[i]) / dc[i];
                }
                auto beta_v = 1.0;
                for (size_t i = 0; i < x.size() - 1; i++) {
                    for (size_t j = i + 1; j < x.size(); j++) {
                        vred_mix1 += 2.0 * x[i] * x[j] * beta_v * gama_v[i][j] * (x[i] + x[j]) / (beta_v * beta_v * x[i] + x[j]) * 0.125 * pow(1.0 / pow(dc[i], 1.0 / 3.0) + 1.0 / pow(dc[j], 1.0 / 3.0), 3.0);
                    }
                }

                return 1.0 / vred_mix1;
            }

            template<typename ETYPE, typename LTYPE, typename MTYPE>
            inline auto get_tc(const ETYPE& e, const LTYPE& l, const MTYPE& m) const {
                return e * (tc_p[0] + tc_p[1] / l + tc_p[2] / (l * l * l)) *
                    (1.0 + 0.5135 * (m - 1.0) * l - 0.0344 * (m - 1.0) * l * l + 0.0037 * (m - 1.0) * (m - 1.0) * (m - 1.0) * l) /
                    (1.0 + 0.3220 * (m - 1.0) * l - 0.0223 * (m - 1.0) * l * l);
            }

            template<typename STYPE, typename LTYPE, typename MTYPE>
            inline auto get_dc(const STYPE& s, const LTYPE& l, const MTYPE& m) const {
                return 1E3 * m * (dc_p[0] + dc_p[1] * log(l) / log(10.0)) *
                    (1.0 - 0.0681 * (m - 1.0) * l + 0.0029 * (m - 1.0) * pow(l, 2.0)) /
                    (1.0 + 0.0462 * (m - 1.0) * l - 0.0019 * (m - 1.0) * pow(l, 2.0)) * 1E27 / (NAvo * pow(s, 3.0)) / m;
            }


            // Approximate the mixture with the one fluid model
            template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
            auto one_fluid(TTYPE& Tstar, RHOTYPE& rhostar, MoleFracType& molefrac) const {

                using resulttype = std::common_type_t<decltype(Tstar), decltype(molefrac[0]), decltype(rhostar)>;
                resulttype l = 0.0;
                resulttype s = 0.0;
                resulttype e = 0.0;
                resulttype m = 0.0;

                std::tie(s, e, l, m) = combining_rules_one_fluid(rhostar, molefrac, fld, is_sphere, k_mat_lambda, k_mat_sigma, k_mat_epsilon, gammat, betat, gamma_dense, mix_rule);

                resulttype tc = get_tc(e, l, m);
                resulttype dc = get_dc(s, l, m);

                resulttype tau = tc / Tstar;
                resulttype delta = rhostar / dc;

                // Calculate coefficients
                std::vector< resulttype> n_pol;
                std::vector< resulttype> n_exp;
                std::vector< resulttype> n_gbs;
                for (size_t i = 0; i < t_pol.size(); i++) { n_pol.push_back(c1_pol[i] + c2_pol[i] / l); }
                for (size_t i = 0; i < t_exp.size(); i++) { n_exp.push_back(c1_exp[i] + c2_exp[i] / l); }
                for (size_t i = 0; i < t_gbs.size(); i++) { n_gbs.push_back(c1_gbs[i] + c2_gbs[i] / l); }


                std::vector< resulttype> n_pol_m;
                std::vector< resulttype> n_exp_m;
                std::vector< resulttype> n_gbs_m;
                for (size_t i = 0; i < t_pol_m.size(); i++) { n_pol_m.push_back(d1_pol[i] + d2_pol[i] * (1.0 - m) + d3_pol[i] * (1.0 - m) * (1.0 - m)); }
                for (size_t i = 0; i < t_exp_m.size(); i++) { n_exp_m.push_back(d1_exp[i] + d2_exp[i] * (1.0 - m) + d3_exp[i] * (1.0 - m) * (1.0 - m)); }
                for (size_t i = 0; i < t_gbs_m.size(); i++) { n_gbs_m.push_back(d1_gbs[i] + d2_gbs[i] * (1.0 - m) + d3_gbs[i] * (1.0 - m) * (1.0 - m)); }

                std::vector< resulttype> pol, exp_, gbs;
                for (size_t i = 0; i < t_pol.size(); i++) { pol.push_back(n_pol[i] * pow(tau, t_pol[i]) * pow(delta, d_pol[i])); }
                for (size_t i = 0; i < t_exp.size(); i++) { exp_.push_back(n_exp[i] * pow(tau, t_exp[i]) * pow(delta, d_exp[i]) * exp(-pow(delta, p[i]))); }
                for (size_t i = 0; i < t_gbs.size(); i++) { gbs.push_back(n_gbs[i] * pow(tau, t_gbs[i]) * pow(delta, d_gbs[i]) * exp(-eta[i] * (delta - eps[i]) * (delta - eps[i]) - beta[i] * (tau - gam[i]) * (tau - gam[i]))); }

                std::vector< resulttype> pol_m, exp_m, gbs_m;
                for (size_t i = 0; i < t_pol_m.size(); i++) { pol_m.push_back(n_pol_m[i] * pow(tau, t_pol_m[i]) * pow(delta, d_pol_m[i])); }
                for (size_t i = 0; i < t_exp_m.size(); i++) { exp_m.push_back(n_exp_m[i] * pow(tau, t_exp_m[i]) * pow(delta, d_exp_m[i]) * exp(-pow(delta, p_m[i]))); }
                for (size_t i = 0; i < t_gbs_m.size(); i++) { gbs_m.push_back(n_gbs_m[i] * pow(tau, t_gbs_m[i]) * pow(delta, d_gbs_m[i]) * exp(-eta_m[i] * (delta - eps_m[i]) * (delta - eps_m[i]) - beta_m[i] * (tau - gam_m[i]) * (tau - gam_m[i]))); }

                auto alpha_r_sphere = std::reduce(pol.begin(), pol.end()) + std::reduce(exp_.begin(), exp_.end()) + std::reduce(gbs.begin(), gbs.end());
                auto alpha_r_el = std::reduce(pol_m.begin(), pol_m.end()) + std::reduce(exp_m.begin(), exp_m.end()) + std::reduce(gbs_m.begin(), gbs_m.end());
                resulttype alpha_r_all = 0.0;
                if (is_sphere) {
                    alpha_r_all = alpha_r_sphere;
                }
                else
                {
                    //alpha_r_all = m * alpha_r_sphere + (1.0 - m) * alpha_r_el;
                    alpha_r_all = alpha_r_sphere + (1.0 - m) * alpha_r_el;
                }

                return alpha_r_all;
            }

            template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
            auto corresponding_states(const TTYPE& Tstar, const RHOTYPE& rhostar, const MoleFracType& molefrac) const {

                using resulttype = std::common_type_t<decltype(Tstar), decltype(molefrac[0]), decltype(rhostar)>;

                auto ncomp = nr_fld;
                // get pure fluid reducing
                std::vector<resulttype> tc(nr_fld, 0.0);
                std::vector<resulttype> dc(nr_fld, 0.0);
                resulttype TC_MIX = 0.0;
                resulttype DC_MIX = 0.0;

                for (size_t i = 0; i < nr_fld; i++) {
                    tc[i] = get_tc(fld[i].epsilons, fld[i].lambdas, fld[i].m);
                    dc[i] = get_dc(fld[i].sigmas, fld[i].lambdas, fld[i].m);
                }

                std::vector<std::vector<resulttype>> gamma_t(ncomp, std::vector<resulttype>(ncomp, 1.0));
                std::vector<std::vector<resulttype>> gamma_v(ncomp, std::vector<resulttype>(ncomp, 1.0));
                if (mix_rule == LORENTZ) {
                    TC_MIX = get_tred_mix(tc, molefrac, gamma_t);
                    DC_MIX = get_vred_mix(dc, molefrac, gamma_v);
                }
                else if (mix_rule == LINEAR) {
                    for (auto i = 0; i < ncomp; i++) {
                        for (auto j = 0; j < ncomp; j++) {
                            auto dom = pow(1.0 / pow(dc[i], 1.0 / 3.0) + 1.0 / pow(dc[j], 1.0 / 3.0), 3.0);
                            auto nom = 1.0 / dc[i] + 1.0 / dc[j];
                            gamma_v[i][j] = 4.0 * nom / dom;
                            gamma_t[i][j] = 0.5 * (tc[i] + tc[j]) / sqrt(tc[i] * tc[j]);
                        };
                    };
                    TC_MIX = get_tred_mix(tc, molefrac, gamma_t);
                    DC_MIX = get_vred_mix(dc, molefrac, gamma_v);
                }

                std::vector<std::vector<resulttype>> n_pol(ncomp, std::vector<resulttype>(t_pol.size(), 0.0));
                std::vector<std::vector<resulttype>> n_exp(ncomp, std::vector<resulttype>(t_exp.size(), 0.0));
                std::vector<std::vector<resulttype>> n_gbs(ncomp, std::vector<resulttype>(t_gbs.size(), 0.0));

                // Calculate coefficients
                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_pol.size(); j++) {
                        n_pol[i][j] = c1_pol[j] + c2_pol[j] / fld[i].lambdas;
                    }
                };
                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_exp.size(); j++) {
                        n_exp[i][j] = c1_exp[j] + c2_exp[j] / fld[i].lambdas;
                    }
                };
                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_gbs.size(); j++) {
                        n_gbs[i][j] = c1_gbs[j] + c2_gbs[j] / fld[i].lambdas;
                    };
                };

                std::vector<std::vector<resulttype>> n_pol_m(ncomp, std::vector<resulttype>(t_pol_m.size(), 0.0));
                std::vector<std::vector<resulttype>> n_exp_m(ncomp, std::vector<resulttype>(t_exp_m.size(), 0.0));
                std::vector<std::vector<resulttype>> n_gbs_m(ncomp, std::vector<resulttype>(t_gbs_m.size(), 0.0));


                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_pol_m.size(); j++) {
                        n_pol_m[i][j] = d1_pol[j] + d2_pol[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) + d3_pol[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) * (1.0 - fld[i].m);
                    };
                };

                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_exp_m.size(); j++) {
                        n_exp_m[i][j] = d1_exp[j] + d2_exp[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) + d3_exp[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) * (1.0 - fld[i].m);
                    };
                };

                for (size_t i = 0; i < ncomp; i++) {
                    for (size_t j = 0; j < t_gbs_m.size(); j++) {
                        n_gbs_m[i][j] = d1_gbs[j] + d2_gbs[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) + d3_gbs[j] * (1.0 - fld[i].m) * (1.0 - fld[i].m) * (1.0 - fld[i].m);
                    };
                };

                std::vector<resulttype> tau;
                std::vector<resulttype> delta;

                for (size_t i = 0; i < ncomp; i++) {
                    tau.push_back(TC_MIX / Tstar);
                    delta.push_back(rhostar / DC_MIX);
                };

                std::vector<resulttype> pol(ncomp, 0.0);
                std::vector<resulttype> exp_(ncomp, 0.0);
                std::vector<resulttype> gbs(ncomp, 0.0);
                std::vector<resulttype> pol_m(ncomp, 0.0);
                std::vector<resulttype> exp_m(ncomp, 0.0);
                std::vector<resulttype> gbs_m(ncomp, 0.0);

                // pure fluid contribution
                for (size_t i = 0; i < ncomp; i++) {
                    pol[i] = 0.0;
                    for (size_t j = 0; j < t_pol.size(); j++) {
                        pol[i] += n_pol[i][j] * pow(tau[i], t_pol[j]) * pow(delta[i], d_pol[j]);
                    };
                };

                for (size_t i = 0; i < ncomp; i++) {
                    exp_[i] = 0.0;
                    for (size_t j = 0; j < t_exp.size(); j++) {
                        exp_[i] += n_exp[i][j] * pow(tau[i], t_exp[j]) * pow(delta[i], d_exp[j]) * exp(-pow(delta[i], p[j]));
                    };
                };

                for (size_t i = 0; i < ncomp; i++) {
                    gbs[i] = 0.0;
                    for (size_t j = 0; j < t_gbs.size(); j++) {
                        gbs[i] += n_gbs[i][j] * pow(tau[i], t_gbs[j]) * pow(delta[i], d_gbs[j]) * exp(-eta[j] * (delta[i] - eps[j]) * (delta[i] - eps[j]) - beta[j] * (tau[i] - gam[j]) * (tau[i] - gam[j]));
                    };
                };


                if (!(is_sphere)) {
                    for (size_t i = 0; i < ncomp; i++) {
                        pol_m[i] = 0.0;
                        for (size_t j = 0; j < t_pol_m.size(); j++) {
                            pol_m[i] += n_pol_m[i][j] * pow(tau[i], t_pol_m[j]) * pow(delta[i], d_pol_m[j]);
                        };
                    };

                    for (size_t i = 0; i < ncomp; i++) {
                        exp_m[i] = 0.0;
                        for (size_t j = 0; j < t_exp_m.size(); j++) {
                            exp_m[i] += n_exp_m[i][j] * pow(tau[i], t_exp_m[j]) * pow(delta[i], d_exp_m[j]) * exp(-pow(delta[i], p_m[j]));
                        };
                    };

                    for (size_t i = 0; i < ncomp; i++) {
                        gbs_m[i] = 0.0;
                        for (size_t j = 0; j < t_gbs_m.size(); j++) {
                            gbs_m[i] += n_gbs_m[i][j] * pow(tau[i], t_gbs_m[j]) * pow(delta[i], d_gbs_m[j]) * exp(-eta_m[j] * (delta[i] - eps_m[j]) * (delta[i] - eps_m[j]) - beta_m[j] * (tau[i] - gam_m[j]) * (tau[i] - gam_m[j]));
                        };
                    };
                }

                // Sum up everything
                resulttype alpha_r = 0.0;
                resulttype alpha_r_sphere = 0.0;
                resulttype alpha_r_elongated = 0.0;

                for (size_t i = 0; i < ncomp; i++) {
                    if (fld[i].is_sphere) {
                        // Single sphere part of component i
                        alpha_r_sphere = pol[i] + exp_[i] + gbs[i];
                        alpha_r += molefrac[i] * alpha_r_sphere; //(fld[i].m * alpha_r_sphere);
                    }
                    else {
                        // Single elongation part of component i
                        alpha_r_sphere = pol[i] + exp_[i] + gbs[i];
                        alpha_r_elongated = pol_m[i] + exp_m[i] + gbs_m[i];
                        alpha_r += molefrac[i] * (alpha_r_sphere + (1.0 - fld[i].m) * alpha_r_elongated);//(fld[i].m * alpha_r_sphere + (1.0 - fld[i].m) * alpha_r_elongated);
                    }
                };

                return alpha_r;
            }



            // Input is temperature in K, density in mol/m^3 and molefractions
            template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
            auto alphar(const TTYPE& Tstar, const RHOTYPE& rhostar, const MoleFracType& molefrac) const {
                using resulttype = std::common_type_t<decltype(Tstar), decltype(molefrac[0]), decltype(rhostar)>;

                resulttype alpha_r_all = 0.0;

                if (mix_rule == ONEFLUID || mix_rule == ONEFLUID_RED || mix_rule == ONEFLUID_RED_LINEAR || mix_rule == ONEFLUID_DENSE) {
                    alpha_r_all = one_fluid(Tstar, rhostar, molefrac);
                }
                else
                {
                    alpha_r_all = corresponding_states(Tstar, rhostar, molefrac);
                }

                return forceeval(alpha_r_all);
            }

        };


    }
}


//    class MieElongMix2 {

//    private:

//        using EArray6 = Eigen::Array<double, 6, 1>;
//        using EArray4 = Eigen::Array<double, 4, 1>;


//        const EArray6 c1_pol = (EArray6() << -0.0192944, 1.38, -2.2653, 1.6291, -1.974, 0.40412).finished();
//        const EArray6 c1_exp = (EArray6() << 0.1845, -0.3227, 1.1351, 2.232, -2.344, -0.4238).finished();
//        const EArray4 c1_gbs = (EArray4() << -4.367, 0.0371, 1.3895, 2.835).finished();
//        const EArray6 c2_pol = (EArray6() << 0.26021, -5.525, 8.329, -19.492, 25.8, -3.8133).finished();
//        const EArray6 c2_exp = (EArray6() << -5.05, 2.7842, -9.523, -30.383, 17.902, 2.2264).finished();
//        const EArray4 c2_gbs = (EArray4() << 48.445, -5.506, -11.643, -24.36).finished();
//        const EArray6 t_pol = (EArray6() << 1, 0.236, 0.872, 0.313, 0.407, 0.703).finished();
//        const EArray6 t_exp = (EArray6() << 1.78, 2.99, 2.866, 1.2, 3.06, 1.073).finished();
//        const EArray4 t_gbs = (EArray4() << 1.50, 1.03, 4.02, 1.57).finished();
//        const EArray6 d_pol = (EArray6() << 4, 1, 1, 2, 2, 3).finished();
//        const EArray6 d_exp = (EArray6() << 1, 1, 3, 2, 2, 5).finished();
//        const EArray4 d_gbs = (EArray4() << 2, 3, 2, 2).finished();
//        const EArray6 p = (EArray6() << 1, 2, 2, 1, 2, 1).finished();
//        const EArray4 eta = (EArray4() << 0.362, 0.313, 1.17, 0.957).finished();
//        const EArray4 beta = (EArray4() << 0.0761, 0.143, 0.63, 1.32).finished();
//        const EArray4 gam = (EArray4() << 1.55, -0.0826, 1.505, 1.07).finished();
//        const EArray4 eps = (EArray4() << -1, -1, -0.195, -0.287).finished();

//        // elongation part
//        const EArray6 d1_pol = (EArray6() << -0.0033717, -0.93963, 1.0403, -0.11078, -0.0025363, 0.0029514).finished();
//        const EArray6 d2_pol = (EArray6() << 0.0056524, -1.0549, 1.2465, -0.22323, 0.15836, 0.11005).finished();
//        const EArray6 d3_pol = (EArray6() << -0.0032422, -0.69384, 0.80325, 0.26957, -0.069081, -0.0026843).finished();
//        const EArray6 d1_exp = (EArray6() << -1.0436, 0.25891, -0.040776, -0.89272, -0.12408, -0.053926).finished();
//        const EArray6 d2_exp = (EArray6() << -0.27931, -0.27319, 0.1812, -1.2567, 0.71416, -0.040811).finished();
//        const EArray6 d3_exp = (EArray6() << -0.26861, -0.032778, 0.17816, -1.3529, 0.58892, 0.0034168).finished();

//        const EArray6 t_pol_m = (EArray6() << 1, 0.684125778408247, 0.431293949327048, 0.686848959577935, 0.39116753396418, 0.755900521521362).finished();
//        const EArray6 t_exp_m = (EArray6() << 0.994097378429697, 0.81534766545966, 1.96973994708541, 1.35262796041634, 3, 1.41415130674305).finished();
//        const EArray6 d_pol_m = (EArray6() << 4, 1, 1, 2, 2, 3).finished();
//        const EArray6 d_exp_m = (EArray6() << 1, 1, 3, 2, 2, 5).finished();
//        const EArray6 p_m = (EArray6() << 1, 2, 2, 1, 2, 1).finished();

//        double m_lambda_a, ms;
//        std::vector<double> lambdas, epsilons, sigmas, ms_s;
//    public:

//        //MieElong(double lambda_a, double ms, double sigma, double eps) { // Repulsive exponent, segment number ms
//        MieElongMix2(const std::string& path, const std::vector<std::string>& fluids) {
//            std::string filepath = std::filesystem::is_regular_file(path) ? path : path;
//            nlohmann::json j = load_a_JSON_file(filepath);

//            for (auto f : fluids)
//            {
//                lambdas.push_back(j.at("mie").at(f).at("lambda"));
//                epsilons.push_back(j.at("mie").at(f).at("epsilon"));
//                sigmas.push_back(j.at("mie").at(f).at("sigma"));
//                ms_s.push_back(j.at("mie").at(f).at("segment"));
//            };

//        };


//        // We are in "simulation units", so R is 1.0, and T and rho that
//        // go into alphar are actually T^* and rho^*
//        template<typename MoleFracType>
//        double R(const MoleFracType&) const { return NAvo * kBoltz; }

//        // Input is temperature in K, density in mol/m^3 and molefractions
//        template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
//        auto alphar(const TTYPE& Tstar, const RHOTYPE& rhostar, const MoleFracType& molefrac) const {

//            auto ncomp = ms_s.size();
//            using resulttype = std::common_type_t<decltype(Tstar), decltype(molefrac[0]), decltype(rhostar)>;


//            std::vector<vector<resulttype>> sigma_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//            std::vector<vector<resulttype>> epsilon_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//            std::vector<vector<resulttype>> lambda_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//            std::vector<vector<resulttype>> gamma_t(ncomp, vector<resulttype>(ncomp, rhostar));
//            std::vector<vector<resulttype>> gamma_v(ncomp, vector<resulttype>(ncomp, rhostar));


//            for (auto i = 0; i < ncomp; i++) {
//                for (auto j = 0; j < ncomp; j++) {
//                    lambda_ij[i][j] = (lambdas[i] + lambdas[j]) / 2.0;
//                    sigma_ij[i][j] = (sigmas[i] + sigmas[j]) / 2.0;
//                    epsilon_ij[i][j] = sqrt(epsilons[i] * epsilons[j]);
//                    auto dom = pow(1.0 / pow(sigmas[i], 1.0 / 3.0) + 1.0 / pow(sigmas[j], 1.0 / 3.0), 3.0);
//                    auto nom = 1.0 / sigmas[i] + 1.0 / sigmas[j];
//                    gamma_v[i][j] = 1.0;//4.0 * nom / dom;
//                    gamma_t[i][j] = 1.0;//0.5 * (epsilons[i] + epsilons[j]) / sqrt(epsilons[i] * epsilons[j]);
//                };
//            };

//            std::vector<resulttype> TC;
//            std::vector<resulttype> DC;
//            for (size_t i = 0; i < lambdas.size(); i++) {
//                TC.push_back(get_tc(epsilons[i], lambdas[i], ms_s[i]));
//                DC.push_back(get_dc(sigmas[i], lambdas[i], ms_s[i]));
//            };


//            std::vector<vector<resulttype>> n_pol(ncomp, vector<resulttype>(t_pol.size(), 0.0));
//            std::vector<vector<resulttype>> n_exp(ncomp, vector<resulttype>(t_exp.size(), 0.0));
//            std::vector<vector<resulttype>> n_gbs(ncomp, vector<resulttype>(t_gbs.size(), 0.0));
//            // Calculate coefficients
//            for (size_t i = 0; i < ncomp; i++) {
//                for (size_t j = 0; j < t_pol.size(); j++) {
//                    n_pol[i][j] = c1_pol[j] + c2_pol[j] / lambdas[i];
//                }
//            };
//            for (size_t i = 0; i < ncomp; i++) {
//                for (size_t j = 0; j < t_exp.size(); j++) {
//                    n_exp[i][j] = c1_exp[j] + c2_exp[j] / lambdas[i];
//                }
//            };
//            for (size_t i = 0; i < ncomp; i++) {
//                for (size_t j = 0; j < t_gbs.size(); j++) {
//                    n_gbs[i][j] = c1_gbs[j] + c2_gbs[j] / lambdas[i];
//                };
//            };

//            std::vector<vector<resulttype>> n_pol_m(ncomp, vector<resulttype>(t_pol_m.size(), 0.0));
//            std::vector<vector<resulttype>> n_exp_m(ncomp, vector<resulttype>(t_exp_m.size(), 0.0));

//            for (size_t i = 0; i < ncomp; i++) {
//                for (size_t j = 0; j < t_pol_m.size(); j++) {
//                    n_pol_m[i][j] = d1_pol[j] + d2_pol[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) + d3_pol[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) * (1.0 - ms_s[i]);
//                };
//            };

//            for (size_t i = 0; i < ncomp; i++) {
//                for (size_t j = 0; j < t_exp_m.size(); j++) {
//                    n_exp_m[i][j] = d1_exp[j] + d2_exp[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) + d3_exp[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) * (1.0 - ms_s[i]);
//                };
//            };

//            std::vector<resulttype> tau;
//            std::vector<resulttype> delta;
//            for (size_t i = 0; i < ncomp; i++) {
//                tau.push_back(get_tred_mix(TC, molefrac, gamma_t) / Tstar);
//                delta.push_back(rhostar / get_vred_mix(DC, molefrac, gamma_v));
//            };

//            std::vector<resulttype> pol(ncomp, 0.0);
//            std::vector<resulttype> exp_(ncomp, 0.0);
//            std::vector<resulttype> gbs(ncomp, 0.0);
//            std::vector<resulttype> pol_m(ncomp, 0.0);
//            std::vector<resulttype> exp_m(ncomp, 0.0);
//            for (size_t i = 0; i < ncomp; i++) {
//                pol[i] = 0.0;
//                for (size_t j = 0; j < t_pol.size(); j++) {
//                    pol[i] += n_pol[i][j] * pow(tau[i], t_pol[j]) * pow(delta[i], d_pol[j]);
//                };
//            };
//            for (size_t i = 0; i < ncomp; i++) {
//                exp_[i] = 0.0;
//                for (size_t j = 0; j < t_exp.size(); j++) {
//                    exp_[i] += n_exp[i][j] * pow(tau[i], t_exp[j]) * pow(delta[i], d_exp[j]) * exp(-pow(delta[i], p[j]));
//                };
//            };
//            for (size_t i = 0; i < ncomp; i++) {
//                gbs[i] = 0.0;
//                for (size_t j = 0; j < t_gbs.size(); j++) {
//                    gbs[i] += n_gbs[i][j] * pow(tau[i], t_gbs[j]) * pow(delta[i], d_gbs[j]) * exp(-eta[j] * (delta[i] - eps[j]) * (delta[i] - eps[j]) - beta[j] * (tau[i] - gam[j]) * (tau[i] - gam[j]));
//                };
//            };

//            for (size_t i = 0; i < ncomp; i++) {
//                pol_m[i] = 0.0;
//                for (size_t j = 0; j < t_pol_m.size(); j++) {
//                    pol_m[i] += n_pol_m[i][j] * pow(tau[i], t_pol_m[j]) * pow(delta[i], d_pol_m[j]);
//                };
//            };
//            for (size_t i = 0; i < ncomp; i++) {
//                exp_m[i] = 0.0;
//                for (size_t j = 0; j < t_exp_m.size(); j++) {
//                    exp_m[i] += n_exp_m[i][j] * pow(tau[i], t_exp_m[j]) * pow(delta[i], d_exp_m[j]) * exp(-pow(delta[i], p_m[j]));
//                };
//            };


//            resulttype alpha_r_sphere = 0.0;
//            resulttype alpha_r_el = 0.0;
//            resulttype alpha_r_all = 0.0;
//            for (size_t i = 0; i < ncomp; i++) {
//                alpha_r_sphere = 0.0;
//                alpha_r_el = 0.0;
//                alpha_r_sphere = pol[i] + exp_[i] + gbs[i];
//                alpha_r_el = pol_m[i] + exp_m[i];
//                alpha_r_all += molefrac[i] * (ms_s[i] * alpha_r_sphere + (1.0 - ms_s[i]) * alpha_r_el);
//            };
//            return forceeval(alpha_r_all);
//        }

//    };

//};


//// new mixing rules try
//class MieElongMix3 {

//private:

//    using EArray6 = Eigen::Array<double, 6, 1>;
//    using EArray4 = Eigen::Array<double, 4, 1>;


//    const EArray6 c1_pol = (EArray6() << -0.0192944, 1.38, -2.2653, 1.6291, -1.974, 0.40412).finished();
//    const EArray6 c1_exp = (EArray6() << 0.1845, -0.3227, 1.1351, 2.232, -2.344, -0.4238).finished();
//    const EArray4 c1_gbs = (EArray4() << -4.367, 0.0371, 1.3895, 2.835).finished();
//    const EArray6 c2_pol = (EArray6() << 0.26021, -5.525, 8.329, -19.492, 25.8, -3.8133).finished();
//    const EArray6 c2_exp = (EArray6() << -5.05, 2.7842, -9.523, -30.383, 17.902, 2.2264).finished();
//    const EArray4 c2_gbs = (EArray4() << 48.445, -5.506, -11.643, -24.36).finished();
//    const EArray6 t_pol = (EArray6() << 1, 0.236, 0.872, 0.313, 0.407, 0.703).finished();
//    const EArray6 t_exp = (EArray6() << 1.78, 2.99, 2.866, 1.2, 3.06, 1.073).finished();
//    const EArray4 t_gbs = (EArray4() << 1.50, 1.03, 4.02, 1.57).finished();
//    const EArray6 d_pol = (EArray6() << 4, 1, 1, 2, 2, 3).finished();
//    const EArray6 d_exp = (EArray6() << 1, 1, 3, 2, 2, 5).finished();
//    const EArray4 d_gbs = (EArray4() << 2, 3, 2, 2).finished();
//    const EArray6 p = (EArray6() << 1, 2, 2, 1, 2, 1).finished();
//    const EArray4 eta = (EArray4() << 0.362, 0.313, 1.17, 0.957).finished();
//    const EArray4 beta = (EArray4() << 0.0761, 0.143, 0.63, 1.32).finished();
//    const EArray4 gam = (EArray4() << 1.55, -0.0826, 1.505, 1.07).finished();
//    const EArray4 eps = (EArray4() << -1, -1, -0.195, -0.287).finished();

//    // elongation part
//    const EArray6 d1_pol = (EArray6() << -0.0033717, -0.93963, 1.0403, -0.11078, -0.0025363, 0.0029514).finished();
//    const EArray6 d2_pol = (EArray6() << 0.0056524, -1.0549, 1.2465, -0.22323, 0.15836, 0.11005).finished();
//    const EArray6 d3_pol = (EArray6() << -0.0032422, -0.69384, 0.80325, 0.26957, -0.069081, -0.0026843).finished();
//    const EArray6 d1_exp = (EArray6() << -1.0436, 0.25891, -0.040776, -0.89272, -0.12408, -0.053926).finished();
//    const EArray6 d2_exp = (EArray6() << -0.27931, -0.27319, 0.1812, -1.2567, 0.71416, -0.040811).finished();
//    const EArray6 d3_exp = (EArray6() << -0.26861, -0.032778, 0.17816, -1.3529, 0.58892, 0.0034168).finished();

//    const EArray6 t_pol_m = (EArray6() << 1, 0.684125778408247, 0.431293949327048, 0.686848959577935, 0.39116753396418, 0.755900521521362).finished();
//    const EArray6 t_exp_m = (EArray6() << 0.994097378429697, 0.81534766545966, 1.96973994708541, 1.35262796041634, 3, 1.41415130674305).finished();
//    const EArray6 d_pol_m = (EArray6() << 4, 1, 1, 2, 2, 3).finished();
//    const EArray6 d_exp_m = (EArray6() << 1, 1, 3, 2, 2, 5).finished();
//    const EArray6 p_m = (EArray6() << 1, 2, 2, 1, 2, 1).finished();

//    double m_lambda_a, ms;
//    std::vector<double> lambdas, epsilons, sigmas, ms_s;
//public:

//    //MieElong(double lambda_a, double ms, double sigma, double eps) { // Repulsive exponent, segment number ms
//    MieElongMix3(const std::string& path, const std::vector<std::string>& fluids) {
//        std::string filepath = std::filesystem::is_regular_file(path) ? path : path;
//        nlohmann::json j = load_a_JSON_file(filepath);

//        for (auto f : fluids)
//        {
//            lambdas.push_back(j.at("mie").at(f).at("lambda"));
//            epsilons.push_back(j.at("mie").at(f).at("epsilon"));
//            sigmas.push_back(j.at("mie").at(f).at("sigma"));
//            ms_s.push_back(j.at("mie").at(f).at("segment"));
//        };

//    };


//    // We are in "simulation units", so R is 1.0, and T and rho that
//    // go into alphar are actually T^* and rho^*
//    template<typename MoleFracType>
//    double R(const MoleFracType&) const { return NAvo * kBoltz; }

//    // Input is temperature in K, density in mol/m^3 and molefractions
//    template<typename TTYPE, typename RHOTYPE, typename MoleFracType>
//    auto alphar(const TTYPE& Tstar, const RHOTYPE& rhostar, const MoleFracType& molefrac) const {

//        auto ncomp = ms_s.size();
//        using resulttype = std::common_type_t<decltype(Tstar), decltype(molefrac[0]), decltype(rhostar)>;


//        std::vector<vector<resulttype>> sigma_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//        std::vector<vector<resulttype>> epsilon_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//        std::vector<vector<resulttype>> lambda_ij(ncomp, vector<resulttype>(ncomp, rhostar));
//        std::vector<vector<resulttype>> gamma_t(ncomp, vector<resulttype>(ncomp, rhostar));
//        std::vector<vector<resulttype>> gamma_v(ncomp, vector<resulttype>(ncomp, rhostar));


//        for (auto i = 0; i < ncomp; i++) {
//            for (auto j = 0; j < ncomp; j++) {
//                lambda_ij[i][j] = (lambdas[i] + lambdas[j]) / 2.0;
//                sigma_ij[i][j] = (sigmas[i] + sigmas[j]) / 2.0;
//                epsilon_ij[i][j] = sqrt(epsilons[i] * epsilons[j]);
//                auto dom = pow(1.0 / pow(sigmas[i], 1.0 / 3.0) + 1.0 / pow(sigmas[j], 1.0 / 3.0), 3.0);
//                auto nom = 1.0 / sigmas[i] + 1.0 / sigmas[j];
//                gamma_v[i][j] = 1.0;//4.0 * nom / dom;
//                gamma_t[i][j] = 1.0;//0.5 * (epsilons[i] + epsilons[j]) / sqrt(epsilons[i] * epsilons[j]);
//            };
//        };

//        std::vector<resulttype> TC;
//        std::vector<resulttype> DC;
//        for (size_t i = 0; i < lambdas.size(); i++) {
//            TC.push_back(get_tc(epsilons[i], lambdas[i], ms_s[i]));
//            DC.push_back(get_dc(sigmas[i], lambdas[i], ms_s[i]));
//        };


//        std::vector<vector<resulttype>> n_pol(ncomp, vector<resulttype>(t_pol.size(), 0.0));
//        std::vector<vector<resulttype>> n_exp(ncomp, vector<resulttype>(t_exp.size(), 0.0));
//        std::vector<vector<resulttype>> n_gbs(ncomp, vector<resulttype>(t_gbs.size(), 0.0));
//        // Calculate coefficients
//        for (size_t i = 0; i < ncomp; i++) {
//            for (size_t j = 0; j < t_pol.size(); j++) {
//                n_pol[i][j] = c1_pol[j] + c2_pol[j] / lambdas[i];
//            }
//        };
//        for (size_t i = 0; i < ncomp; i++) {
//            for (size_t j = 0; j < t_exp.size(); j++) {
//                n_exp[i][j] = c1_exp[j] + c2_exp[j] / lambdas[i];
//            }
//        };
//        for (size_t i = 0; i < ncomp; i++) {
//            for (size_t j = 0; j < t_gbs.size(); j++) {
//                n_gbs[i][j] = c1_gbs[j] + c2_gbs[j] / lambdas[i];
//            };
//        };

//        std::vector<vector<resulttype>> n_pol_m(ncomp, vector<resulttype>(t_pol_m.size(), 0.0));
//        std::vector<vector<resulttype>> n_exp_m(ncomp, vector<resulttype>(t_exp_m.size(), 0.0));

//        for (size_t i = 0; i < ncomp; i++) {
//            for (size_t j = 0; j < t_pol_m.size(); j++) {
//                n_pol_m[i][j] = d1_pol[j] + d2_pol[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) + d3_pol[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) * (1.0 - ms_s[i]);
//            };
//        };

//        for (size_t i = 0; i < ncomp; i++) {
//            for (size_t j = 0; j < t_exp_m.size(); j++) {
//                n_exp_m[i][j] = d1_exp[j] + d2_exp[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) + d3_exp[j] * (1.0 - ms_s[i]) * (1.0 - ms_s[i]) * (1.0 - ms_s[i]);
//            };
//        };

//        std::vector<resulttype> tau;
//        std::vector<resulttype> delta;
//        /*for (size_t i = 0; i < ncomp; i++) {
//            tau.push_back(TC[i] / Tstar);
//            delta.push_back(rhostar / DC[i]);
//        };*/
//        for (size_t i = 0; i < ncomp; i++) {
//            tau.push_back(TC[i] / Tstar);
//            delta.push_back(rhostar / DC[i]);
//        };

//        std::vector<resulttype> pol(ncomp, 0.0);
//        std::vector<resulttype> exp_(ncomp, 0.0);
//        std::vector<resulttype> gbs(ncomp, 0.0);
//        std::vector<resulttype> pol_m(ncomp, 0.0);
//        std::vector<resulttype> exp_m(ncomp, 0.0);

//        std::vector<resulttype> pol_dep(ncomp - 1, 0.0);
//        std::vector<resulttype> exp_dep(ncomp - 1, 0.0);
//        std::vector<resulttype> gbs_dep(ncomp - 1, 0.0);
//        std::vector<resulttype> pol_m_dep(ncomp - 1, 0.0);
//        std::vector<resulttype> exp_m_dep(ncomp - 1, 0.0);


//        // pure fluid contribution
//        for (size_t i = 0; i < ncomp; i++) {
//            pol[i] = 0.0;
//            for (size_t j = 0; j < t_pol.size(); j++) {
//                pol[i] += n_pol[i][j] * pow(tau[i], t_pol[j]) * pow(delta[i], d_pol[j]);
//            };
//        };

//        for (size_t i = 0; i < ncomp; i++) {
//            exp_[i] = 0.0;
//            for (size_t j = 0; j < t_exp.size(); j++) {
//                exp_[i] += n_exp[i][j] * pow(tau[i], t_exp[j]) * pow(delta[i], d_exp[j]) * exp(-pow(delta[i], p[j]));
//            };
//        };

//        for (size_t i = 0; i < ncomp; i++) {
//            gbs[i] = 0.0;
//            for (size_t j = 0; j < t_gbs.size(); j++) {
//                gbs[i] += n_gbs[i][j] * pow(tau[i], t_gbs[j]) * pow(delta[i], d_gbs[j]) * exp(-eta[j] * (delta[i] - eps[j]) * (delta[i] - eps[j]) - beta[j] * (tau[i] - gam[j]) * (tau[i] - gam[j]));
//            };
//        };

//        for (size_t i = 0; i < ncomp; i++) {
//            pol_m[i] = 0.0;
//            for (size_t j = 0; j < t_pol_m.size(); j++) {
//                pol_m[i] += n_pol_m[i][j] * pow(tau[i], t_pol_m[j]) * pow(delta[i], d_pol_m[j]);
//            };
//        };

//        for (size_t i = 0; i < ncomp; i++) {
//            exp_m[i] = 0.0;
//            for (size_t j = 0; j < t_exp_m.size(); j++) {
//                exp_m[i] += n_exp_m[i][j] * pow(tau[i], t_exp_m[j]) * pow(delta[i], d_exp_m[j]) * exp(-pow(delta[i], p_m[j]));
//            };
//        };

//        // depature contribution
//        /*for (size_t i = 0; i < ncomp - 1; i++) {
//            pol_dep[i] = 0.0;
//            for (size_t j = 0; j < t_pol.size(); j++) {
//                pol_dep[i] += (n_pol[i][j] + n_pol[i + 1][j]) / 2.0 * pow(tau[i], t_pol[j]) * pow(delta[i], d_pol[j]);
//            };
//        };

//        for (size_t i = 0; i < ncomp - 1; i++) {
//            exp_dep[i] = 0.0;
//            for (size_t j = 0; j < t_exp.size(); j++) {
//                exp_dep[i] += (n_exp[i][j] + n_exp[i + 1][j]) / 2.0 * pow(tau[i], t_exp[j]) * pow(delta[i], d_exp[j]) * exp(-pow(delta[i], p[j]));
//            };
//        };

//        for (size_t i = 0; i < ncomp - 1; i++) {
//            gbs_dep[i] = 0.0;
//            for (size_t j = 0; j < t_gbs.size(); j++) {
//                gbs_dep[i] += (n_gbs[i][j] + n_gbs[i + 1][j]) / 2.0 * pow(tau[i], t_gbs[j]) * pow(delta[i], d_gbs[j]) * exp(-eta[j] * (delta[i] - eps[j]) * (delta[i] - eps[j]) - beta[j] * (tau[i] - gam[j]) * (tau[i] - gam[j]));
//            };
//        };

//        for (size_t i = 0; i < ncomp - 1; i++) {
//            pol_m_dep[i] = 0.0;
//            for (size_t j = 0; j < t_pol_m.size(); j++) {
//                pol_m_dep[i] += (n_pol_m[i][j] + n_pol_m[i + 1][j]) / 2.0 * pow(tau[i], t_pol_m[j]) * pow(delta[i], d_pol_m[j]);
//            };
//        };
//        for (size_t i = 0; i < ncomp - 1; i++) {
//            exp_m_dep[i] = 0.0;
//            for (size_t j = 0; j < t_exp_m.size(); j++) {
//                exp_m_dep[i] += (n_exp_m[i][j] + n_exp_m[i + 1][j]) / 2.0 * pow(tau[i], t_exp_m[j]) * pow(delta[i], d_exp_m[j]) * exp(-pow(delta[i], p_m[j]));
//            };
//        };*/
//        resulttype m_mix = 0.0;
//        for (size_t i = 0; i < ncomp; i++) {
//            m_mix += ms_s[i] * molefrac[i];
//        }

//        resulttype alpha_r_sphere = 0.0;
//        resulttype alpha_r_el = 0.0;
//        resulttype alpha_r_all = 0.0;

//        std::vector<resulttype> sphere_part(ncomp, 0.0);
//        std::vector<resulttype> elongation_part(ncomp, 0.0);
//        resulttype sphere_mix = 0.0;
//        resulttype elongation_mix = 0.0;
//        resulttype cross_mix = 0.0;

//        for (size_t i = 0; i < ncomp; i++) {
//            // Single sphere part of component i
//            sphere_part[i] = pol[i] + exp_[i] + gbs[i];

//            // Single elongation part of component j
//            elongation_part[i] = pol_m[i] + exp_m[i];

//            sphere_mix += ms_s[i] * molefrac[i] * sphere_part[i];

//        };

//        for (size_t i = 0; i < ncomp - 1; i++) {
//            for (size_t j = i + 1; j < ncomp; j++) {
//                /*cross_mix += (molefrac[i] * molefrac[j] * ms_s[i]*(1.0 - ms_s[j])*sphere_part[i]*elongation_part[j] +
//                              ms_s[j]*(1.0 - ms_s[i])*sphere_part[j]*elongation_part[i] * molefrac[i] * molefrac[j]);*/
//                elongation_mix = (1.0 - ms_s[i]) * molefrac[i] * elongation_part[i] + (1.0 - ms_s[j]) * molefrac[j] * elongation_part[j];
//            }
//        }

//        alpha_r_all = sphere_mix + elongation_mix + cross_mix;
//        return forceeval(alpha_r_all);
//    }

//};


